REM
REM    * * * * * * * * * * * * * * * * * * * * * * * * * * *
REM    *  Grumpy, the most grouchy http server out there!  *
REM    *  Copyright (C) Mateusz Viste 2008                 *
REM    * * * * * * * * * * * * * * * * * * * * * * * * * * *
REM
REM  Written in FreeBASIC v0.20.0 by Mateusz Viste <mateusz.viste@mail.ru>
REM
REM   ----------------------------------------------------------------------
REM    This program is free software: you can redistribute it and/or modify
REM    it under the terms of the GNU General Public License as published by
REM    the Free Software Foundation, either version 3 of the License, or
REM    (at your option) any later version.
REM
REM    This program is distributed in the hope that it will be useful,
REM    but WITHOUT ANY WARRANTY; without even the implied warranty of
REM    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
REM    GNU General Public License for more details.
REM
REM    You should have received a copy of the GNU General Public License
REM    along with this program.  If not, see <http://www.gnu.org/licenses/>.
REM   ----------------------------------------------------------------------
REM

#INCLUDE ONCE "vbcompat.bi" ' Required for use of FileExists(), FORMAT() and FileDateTime() functions
#INCLUDE ONCE "crt.bi"      ' Needed by GetSrvTime
#INCLUDE ONCE "base64.bi"   ' Allows to use the EncBase64() function

DECLARE SUB ReturnFile(FileToReturn AS STRING, HeadQuery AS BYTE = 0)
DECLARE SUB Return301(RedirectTo AS STRING, HeadQuery AS BYTE = 0)
DECLARE SUB Return400(HeadQuery AS BYTE = 0)
DECLARE SUB Return401(HeadQuery AS BYTE = 0, AuthRealm AS STRING)
DECLARE SUB Return403(HeadQuery AS BYTE = 0)
DECLARE SUB Return404(MissingFile AS STRING, HeadQuery AS BYTE = 0)
DECLARE SUB Return500(HeadQuery AS BYTE = 0)
DECLARE SUB Return501(HeadQuery AS BYTE = 0)
DECLARE SUB Return505(HeadQuery AS BYTE = 0)
DECLARE SUB ListDir(DirectoryToList AS STRING, HeadQuery AS BYTE = 0)
DECLARE SUB About()
DECLARE FUNCTION RemoveDoubleSlash(StringToRemove AS STRING) AS STRING
DECLARE FUNCTION ReadCFG(CFGfile AS STRING, CFGField AS STRING) AS STRING
DECLARE FUNCTION CheckForEvasion(JailPath AS STRING, EvadingFile AS STRING) AS BYTE
DECLARE FUNCTION HttpSecurityChecks(Query AS STRING) AS STRING
DECLARE FUNCTION IsDirectory(ElementToCheck AS STRING) AS BYTE
DECLARE FUNCTION GetSrvTime() AS STRING
DECLARE FUNCTION CheckForAuth(PathToCheckFor AS STRING) AS STRING
DECLARE FUNCTION SortStrings(BYVAL StringA AS STRING, BYVAL StringB AS STRING) AS BYTE
DECLARE FUNCTION TranslatePercentEnc(RawString AS STRING, RevEnc AS BYTE = 0) AS STRING
DECLARE FUNCTION LineInput(SocketToListenTo AS INTEGER = 0) AS STRING
DECLARE FUNCTION GetFileSize(FileToCheck AS STRING) AS STRING
DECLARE FUNCTION GetFileDate(FileToCheck AS STRING) AS STRING
DECLARE FUNCTION GetFileDescription(FileToCheck AS STRING, DescrPath AS STRING) AS STRING

CONST pVer AS STRING = "0.11"
CONST pDate AS STRING = "2008"

' Initializing buffers:
DIM SHARED AS STRING LineBuff, OperationBuff, HostBuff, UserAgentBuff, AcceptBuff, AcceptLanguageBuff,_
              AcceptEncodingBuff, AcceptCharsetBuff, KeepAliveBuff, ConnectionBuff, CookieBuff,_
              RefererBuff, IfModifiedSinceBuff, IfUnmodifiedSinceBuff, AuthorizationBuff
DIM SHARED AS STRING GetQuery, GetParam1, GetParam2
' Other variables:
DIM SHARED AS BYTE VerboseMode, AllowDirListing
DIM SHARED AS STRING Token, Value, RootPath, LocalFile, SecurityComment, TempString
DIM HeadFlag AS BYTE

VerboseMode = VAL(ReadCFG("/etc/grumpy.cfg", "Verbose"))
AllowDirListing = VAL(ReadCFG("/etc/grumpy.cfg", "AllowDirListing"))
RootPath = ReadCFG("/etc/grumpy.cfg", "RootDir")
IF RootPath = "" THEN
  RootPath = "/var/www/"
  VerboseMode = 1
  AllowDirListing = 0
END IF
IF RIGHT(RootPath, 1) <> "/" THEN RootPath += "/"

IF LCASE(COMMAND(1)) = "-h" OR LCASE(COMMAND(1)) = "--help" THEN About()

OPEN CONS FOR INPUT AS #4
OPEN CONS FOR OUTPUT AS #5
IF VerboseMode > 0 THEN
  OPEN "/var/log/grumpy.log" FOR APPEND AS #10
  PRINT #10, "" : PRINT #10, "---[ "; Date ; " / "; time; " ]---"
END IF

DO
  LineBuff = LineInput(4)
  IF VerboseMode = 2 THEN PRINT #10, LineBuff

  REM * Perform sanity checks on the request...
  SecurityComment = HttpSecurityChecks(LineBuff)
  IF SecurityComment <> "" THEN
    IF VerboseMode > 0 THEN
      PRINT #10, "***** SECURITY CHECKS *****"
      PRINT #10, " The security module has detected a problem with the following request:"
      PRINT #10, " """; LineBuff; """"
      PRINT #10, " The request has been killed. You will find a short explanation below."
      PRINT #10, " "; CHR(34); SecurityComment; CHR(34)
      PRINT #10, "***************************"
      CLOSE #10
    END IF
    CLOSE #4
    CLOSE #5
    END
  END IF
  REM * Sanity checks done

  Token = UCASE(LEFT(LineBuff, INSTR(LineBuff, " ") - 1))
  Value = TRIM(MID(LineBuff, INSTR(LineBuff, " ") + 1))
  SELECT CASE Token
    CASE "GET", "HEAD", "POST", "PUT", "DELETE", "CONNECT", "TRACE", "OPTIONS"
      OperationBuff = LineBuff
    CASE "HOST:"
      HostBuff = Value
    CASE "USER-AGENT:"
      UserAgentBuff = Value
    CASE "ACCEPT:"
      AcceptBuff = Value
    CASE "ACCEPT-LANGUAGE:"
      AcceptLanguageBuff = Value
    CASE "ACCEPT-ENCODING:"
      AcceptEncodingBuff = Value
    CASE "ACCEPT-CHARSET:"
      AcceptCharsetBuff = Value
    CASE "KEEP-ALIVE:"
      KeepAliveBuff = Value
    CASE "CONNECTION:"
      ConnectionBuff = Value
    CASE "REFERER:"
      RefererBuff = Value
    CASE "COOKIE:"
      CookieBuff = Value
    CASE "IF-MODIFIED-SINCE:"
      IfModifiedSinceBuff = Value
    CASE "IF-UNMODIFIED-SINCE:"
      IfUnmodifiedSinceBuff = Value
    CASE "AUTHORIZATION:"
      AuthorizationBuff = Value
  END SELECT
LOOP UNTIL LEN(LineBuff) = 0

REM * Check for blacklisted user-agents
REM WRITE ME!

REM * Process the request...

GetQuery = MID(OperationBuff, 1, INSTR(OperationBuff, " ") - 1)
GetParam1 = MID(MID(OperationBuff, LEN(GetQuery) + 2), 1, INSTR(MID(OperationBuff, LEN(GetQuery) + 2), " ") - 1)
GetParam2 = MID(OperationBuff, LEN(GetQuery) + 3 + LEN(GetParam1))

GetQuery = TranslatePercentEnc(GetQuery)            '
GetParam1 = TranslatePercentEnc(GetParam1)          ' Decode % encoding before further treatement
GetParam2 = TranslatePercentEnc(GetParam2)          '

IF VerboseMode > 0 THEN PRINT #10, "Query="; GetQuery; " Param1="; GetParam1; " Param2="; GetParam2

SELECT CASE GetQuery
  CASE "GET", "HEAD"
    IF GetQuery = "HEAD" THEN HeadFlag = 1 ELSE HeadFlag = 0
    SELECT CASE GetParam2
      CASE "HTTP/1.0", "HTTP/1.1"
        LocalFile = RemoveDoubleSlash(RootPath + "/" + GetParam1)
        IF CheckForEvasion(RootPath, LocalFile) = 1 THEN
            Return403(HeadFlag)
          ELSE
            IF IsDirectory(LocalFile) = 1 THEN
                IF RIGHT(LocalFile, 1) = "/" THEN
                    TempString = CheckForAuth(LocalFile)
                    IF TempString <> "" THEN
                        Return401(HeadFlag, TempString)
                      ELSE
                        IF FileExists(LocalFile + "index.htm") <> 0 THEN
                            LocalFile += "index.htm"
                            ReturnFile(LocalFile, HeadFlag)
                          ELSEIF FileExists(LocalFile + "index.html") <> 0 THEN
                            LocalFile += "index.html"
                            ReturnFile(LocalFile, HeadFlag)
                          ELSEIF AllowDirListing = 1 THEN
                            ListDir(LocalFile, HeadFlag)
                          ELSE
                            Return403(HeadFlag)
                        END IF
                    END IF
                  ELSE
                    Return301(GetParam1 + "/", HeadFlag)
                END IF
              ELSE
                TempString = CheckForAuth(LocalFile)
                IF TempString <> "" THEN
                    Return401(HeadFlag, TempString)
                  ELSE
                    IF FileExists(LocalFile) <> 0 AND RIGHT(LocalFile, 13) <> "/.grumpy.auth" THEN
                        ReturnFile(LocalFile, HeadFlag)
                      ELSE
                        Return404(GetParam1, HeadFlag)
                    END IF
                END IF
            END IF
        END IF
      CASE ELSE
        Return505(HeadFlag)
    END SELECT
  CASE "POST", "PUT", "DELETE", "CONNECT", "TRACE", "OPTIONS"
    Return501()
  CASE ELSE
    Return400()
END SELECT

CLOSE #4
CLOSE #5
IF VerboseMode > 0 THEN CLOSE #10


REM ***  END OF THE MAIN PROGRAM  ***


SUB Return404(MissingFile AS STRING, HeadFlag AS BYTE = 0)
  IF VerboseMode > 0 THEN PRINT #10, "Response: 404 Not found ("; MissingFile; ")"
  PRINT #5, "HTTP/1.1 404 Not found"
  PRINT #5, "Date: "; GetSrvTime()
  PRINT #5, "Server: Grumpy/"; pVer
  PRINT #5, "Connection: close"
  PRINT #5, "Allow: GET, HEAD"
  PRINT #5, "Accept-Ranges: none"
  'PRINT #5, "Content-Length: 1030"
  PRINT #5, "Content-Type: text/html;charset=utf-8"
  PRINT #5,
  IF HeadFlag = 0 THEN PRINT #5, "<html><head><title>404</title></head><body><h1>404 - Not found</h1></body></html>"
END SUB


SUB ReturnFile(FileToReturn AS STRING, HeadFlag AS BYTE = 0)
  DIM AS INTEGER x, FileLength
  DIM AS STRING*1 ByteBuff
  IF VerboseMode > 0 THEN PRINT #10, "Response: 200 OK ("; FileToReturn; ")"

  OPEN FileToReturn FOR BINARY AS #20
  PRINT #5, "HTTP/1.1 200 OK"
  PRINT #5, "Date: "; GetSrvTime()
  PRINT #5, "Server: Grumpy/"; pVer
  PRINT #5, "Connection: close"
  PRINT #5, "Allow: GET, HEAD"
  PRINT #5, "Accept-Ranges: none"
  PRINT #5, "Content-Length: " & LOF(20)
  'PRINT #5, "Content-Type: text/html; charset=UTF-8"
  PRINT #5,

  IF HeadFlag = 0 THEN   ' Don't return the file content for HEAD queries
    FOR x = 1 TO LOF(20)
      GET #20, x, ByteBuff
      PRINT #5, ByteBuff;
    NEXT x
  END IF
  CLOSE #20
END SUB


SUB Return401(HeadFlag AS BYTE = 0, AuthRealm AS STRING)
  IF VerboseMode > 0 THEN PRINT #10, "Response: 401 Unauthorized"
  PRINT #5, "HTTP/1.1 401 Unauthorized"
  PRINT #5, "Date: "; GetSrvTime()
  PRINT #5, "Server: Grumpy/"; pVer
  PRINT #5, "Connection: close"
  PRINT #5, "Allow: GET, HEAD"
  PRINT #5, "Accept-Ranges: none"
  'PRINT #5, "Content-Length: 8700"
  PRINT #5, "Content-Type: text/html; charset=UTF-8"
  PRINT #5, "WWW-Authenticate: Basic realm="; CHR(34); AuthRealm; CHR(34)
  PRINT #5,
  IF HeadFlag = 0 THEN PRINT #5, "<html><head><title>401</title></head><body><h1>401 - Unauthorized</h1></body></html>"
END SUB


SUB Return403(HeadFlag AS BYTE = 0)
  IF VerboseMode > 0 THEN PRINT #10, "Response: 403 Forbidden"
  PRINT #5, "HTTP/1.1 403 Forbidden"
  PRINT #5, "Date: "; GetSrvTime()
  PRINT #5, "Server: Grumpy/"; pVer
  PRINT #5, "Connection: close"
  PRINT #5, "Allow: GET, HEAD"
  PRINT #5, "Accept-Ranges: none"
  'PRINT #5, "Content-Length: 8700"
  PRINT #5, "Content-Type: text/html; charset=UTF-8"
  PRINT #5,
  IF HeadFlag = 0 THEN PRINT #5, "<html><head><title>403</title></head><body><h1>403 - Forbidden</h1></body></html>"
END SUB


SUB Return501(HeadFlag AS BYTE = 0)
  IF VerboseMode > 0 THEN PRINT #10, "Response: 501 Not implemented"
  PRINT #5, "HTTP/1.1 501 Not implemented"
  PRINT #5, "Date: "; GetSrvTime()
  PRINT #5, "Server: Grumpy/"; pVer
  PRINT #5, "Connection: close"
  PRINT #5, "Allow: GET, HEAD"
  PRINT #5, "Accept-Ranges: none"
  'PRINT #5, "Content-Length: 8700"
  PRINT #5, "Content-Type: text/html; charset=UTF-8"
  PRINT #5,
  IF HeadFlag = 0 THEN PRINT #5, "<html><head><title>501</title></head><body><h1>501 - Not implemented</h1></body></html>"
END SUB


SUB Return400(HeadFlag AS BYTE = 0)
  IF VerboseMode > 0 THEN PRINT #10, "Response: 400 Bad request"
  PRINT #5, "HTTP/1.1 400 Bad request"
  PRINT #5, "Date: "; GetSrvTime()
  PRINT #5, "Server: Grumpy/"; pVer
  PRINT #5, "Connection: close"
  PRINT #5, "Allow: GET, HEAD"
  PRINT #5, "Accept-Ranges: none"
  'PRINT #5, "Content-Length: 8700"
  PRINT #5, "Content-Type: text/html; charset=UTF-8"
  PRINT #5,
  IF HeadFlag = 0 THEN PRINT #5, "<html><head><title>400</title></head><body><h1>400 - Bad request</h1></body></html>"
END SUB


SUB Return505(HeadFlag AS BYTE = 0)
  IF VerboseMode > 0 THEN PRINT #10, "Response: 505 HTTP version not supported"
  PRINT #5, "HTTP/1.1 505 HTTP version not supported"
  PRINT #5, "Date: "; GetSrvTime()
  PRINT #5, "Server: Grumpy/"; pVer
  PRINT #5, "Connection: close"
  PRINT #5, "Allow: GET, HEAD"
  PRINT #5, "Accept-Ranges: none"
  'PRINT #5, "Content-Length: 8700"
  PRINT #5, "Content-Type: text/html; charset=UTF-8"
  PRINT #5,
  IF HeadFlag = 0 THEN PRINT #5, "<html><head><title>505</title></head><body><h1>505 - HTTP version not supported</h1></body></html>"
END SUB


SUB Return500(HeadFlag AS BYTE = 0)
  IF VerboseMode > 0 THEN PRINT #10, "Response: 500 Internal server error"
  PRINT #5, "HTTP/1.1 500 Internal server error"
  PRINT #5, "Date: "; GetSrvTime()
  PRINT #5, "Server: Grumpy/"; pVer
  PRINT #5, "Connection: close"
  PRINT #5, "Allow: GET, HEAD"
  PRINT #5, "Accept-Ranges: none"
  'PRINT #5, "Content-Length: 8700"
  PRINT #5, "Content-Type: text/html; charset=UTF-8"
  PRINT #5,
  IF HeadFlag = 0 THEN PRINT #5, "<html><head><title>500</title></head><body><h1>500 - Internal server error</h1></body></html>"
END SUB


SUB ListDir(DirectoryToList AS STRING, HeadFlag AS BYTE = 0)
  DIM AS STRING FileName, ListTitle, EntryColor, EntrySize, EntryDescription
  DIM AS UINTEGER FileAttrib
  DIM AS INTEGER DirIndex, x, y
  DIM DirList(1 TO 10000, 1 TO 2) AS STRING
  IF VerboseMode > 0 THEN PRINT #10, "Response: 200 OK (list directory """; DirectoryToList; """)"
  ListTitle = MID(DirectoryToList, LEN(RootPath))
  PRINT #5, "HTTP/1.1 200 OK"
  PRINT #5, "Date: "; GetSrvTime()
  PRINT #5, "Server: Grumpy/"; pVer
  PRINT #5, "Connection: close"
  PRINT #5, "Allow: GET, HEAD"
  PRINT #5, "Accept-Ranges: none"
  'PRINT #5, "Content-Length: 8700"
  PRINT #5, "Content-Type: text/html; charset=UTF-8"
  PRINT #5,
  IF HeadFlag = 0 THEN
    PRINT #5, "<!DOCTYPE HTML PUBLIC ""-//W3C//DTD HTML 4.01 Transitional//EN"" ""http://www.w3.org/TR/html4/loose.dtd"">"
    PRINT #5, "<html><head><title>Index of "; ListTitle; "</title></head><body bgcolor=""#DADAFF""><br><table cellpadding=""0"" cellspacing=""0"" width=""100%"" bgcolor=""#EAEAFF"" style=""border: 1px solid #000000""><tr><td><b><big>&nbsp;http://"; HostBuff; ListTitle; "</big></b><br>&nbsp;<i>"; GetFileDescription(".", DirectoryToList); "</i>"
    PRINT #5, "<table cellpadding=""2"" cellspacing=""0"" width=""100%"" border=""0"">"

    FileName = DIR(DirectoryToList + "*", &h37, FileAttrib)
    DirIndex = 0
    DO
      IF FileName <> "." AND (FileName <> ".." OR DirectoryToList <> RootPath) AND (LEFT(FileName, 1) <> "." OR FileName = "..") AND Filename <> "descript.ion" THEN
        DirIndex += 1
        DirList(DirIndex, 1) = FileName
        IF FileAttrib AND &h10 THEN
            DirList(DirIndex, 2) = CHR(9) + "DIR"
          ELSE
            DirList(DirIndex, 2) = ""
        END IF
      END IF
      FileName = DIR(FileAttrib)
    LOOP UNTIL FileName = "" OR DirIndex = 10000

    ' Here I am sorting the directory's entries
    FOR x = 1 TO DirIndex
      FOR y = 1 TO DirIndex
        IF SortStrings(DirList(x, 2) + LCASE(DirList(x, 1)), DirList(y, 2) + LCASE(DirList(y, 1))) = -1 THEN
          Swap DirList(x, 1), DirList(y, 1)
          Swap DirList(x, 2), DirList(y, 2)
        END IF
      NEXT y
    NEXT x

    ' Now, we will display all entries
    EntryColor = """#C2C2EF"""
    PRINT #5, "<tr><td bgcolor="; EntryColor; ">&nbsp;<b>Name</b></td>"
    PRINT #5, "<td bgcolor="; EntryColor; ">&nbsp;<b>Size</b></td>"
    PRINT #5, "<td bgcolor="; EntryColor; ">&nbsp;<b>Date / Time</b></td>"
    PRINT #5, "<td bgcolor="; EntryColor; ">&nbsp;<b>Description</b></td></tr>"

    FOR x = 1 TO DirIndex
      IF x MOD 2 = 0 THEN EntryColor = """#E1E1FA""" ELSE EntryColor = """#EAEAFF"""
      IF LEN(DirList(x, 2)) > 0 THEN
          PRINT #5, "<tr><td bgcolor="; EntryColor; ">&nbsp;<a href="""; TranslatePercentEnc(DirList(x, 1), 1); "/"">"; DirList(x, 1); "</a></td>"
          PRINT #5, "<td bgcolor="; EntryColor; ">&nbsp;[DIR]</td>"
          PRINT #5, "<td bgcolor="; EntryColor; ">&nbsp;"; GetFileDate(DirectoryToList + DirList(x, 1)); "</td>"
          PRINT #5, "<td bgcolor="; EntryColor; ">&nbsp;"; GetFileDescription(DirList(x, 1), DirectoryToList); "</td></tr>"
        ELSE
          PRINT #5, "<tr><td bgcolor="; EntryColor; ">&nbsp;<a href="""; TranslatePercentEnc(DirList(x, 1), 1); """ target=""_blank"">"; DirList(x, 1); "</a></td>"
          PRINT #5, "<td bgcolor="; EntryColor; ">&nbsp;"; GetFileSize(DirectoryToList + DirList(x, 1)); "</td>"
          PRINT #5, "<td bgcolor="; EntryColor; ">&nbsp;"; GetFileDate(DirectoryToList + DirList(x, 1)); "</td>"
          PRINT #5, "<td bgcolor="; EntryColor; ">&nbsp;"; GetFileDescription(DirList(x, 1), DirectoryToList); "</td></tr>"
      END IF
    NEXT x

    PRINT #5, "</table></td></tr></table>"
    PRINT #5, "<div align=""right"" style=""color:#404040;""><small><small>Grumpy HTTP server v"; pVer; " Copyright &copy; Mateusz Viste "; pDate; "</small></small></div>"
    PRINT #5, "</body></html>"
  END IF
END SUB


FUNCTION RemoveDoubleSlash(StringToCheck AS STRING) AS STRING
  DIM AS STRING Wynik
  DIM AS INTEGER x
  Wynik = StringToCheck
  DO
    x = INSTR(Wynik, "//")
    IF x > 0 THEN Wynik = MID(Wynik, 1, x) + MID(Wynik, x + 2)
  LOOP UNTIL INSTR(Wynik, "//") = 0
  RETURN Wynik
END FUNCTION


FUNCTION ReadCFG(CFGfile AS STRING, CFGField AS STRING) AS STRING
 STATIC CfgTable(1 TO 2, 0 TO 255) AS STRING
 DIM AS STRING CfgReturnString, CfgTmpBuffer
 DIM AS INTEGER Counter

 IF CfgTable(1, 0) <> "init ok" THEN
   DIM AS INTEGER CfgFileHandler, CfgColonPos, Counter
   CfgFileHandler = FREEFILE
   CfgReturnString = ""
   Counter = 0
   CfgTable(1, 0) = "init ok"
   IF DIR(CFGfile) <> "" THEN
      OPEN CFGfile FOR INPUT AS #CfgFileHandler
      DO
         Counter += 1
         LINE INPUT #CfgFileHandler, CfgTmpBuffer
         IF MID(TRIM(CfgTmpBuffer), 1, 1) <> "#" THEN
           CfgColonPos = INSTR(CfgTmpBuffer, "=")
           CfgTable(1, Counter) = TRIM(MID(CfgTmpBuffer, 1, CfgColonPos - 1))
           CfgTable(2, Counter) = TRIM(MID(CfgTmpBuffer, CfgColonPos + 1))
         END IF
      LOOP UNTIL EOF(CfgFileHandler) OR Counter = 255
      CLOSE #CfgFileHandler
   END IF
   CfgTable(2, 0) = STR(Counter)
 END IF

 Counter = 0
 DO
   Counter += 1
   IF UCASE(CfgTable(1, Counter)) = UCASE(CFGField) THEN CfgReturnString = CfgTable(2, Counter)
 LOOP UNTIL CfgReturnString <> "" OR Counter >= VAL(CfgTable(2, 0))

 RETURN CfgReturnString
END FUNCTION


FUNCTION CheckForEvasion(JailPath AS STRING, EvadingFile AS STRING) AS BYTE
  DIM AS BYTE Wynik
  IF INSTR(EvadingFile, "..") > 0 THEN Wynik = 1 ELSE Wynik = 0
  RETURN Wynik
END FUNCTION


FUNCTION IsDirectory(ElementToCheck AS STRING) AS BYTE
  DIM AS BYTE Wynik
  DIM AS UINTEGER GetAttribs
  DIM AS STRING TempString
  TempString = DIR(ElementToCheck, &h37, GetAttribs)
  IF GetAttribs AND &h10 THEN Wynik = 1 ELSE Wynik = 0
  Return Wynik
END FUNCTION


SUB Return301(RedirectTo AS STRING, HeadFlag AS BYTE = 0)
  IF VerboseMode > 0 THEN PRINT #10, "Response: 301 ("; RedirectTo; ")"
  PRINT #5, "HTTP/1.1 301"
  PRINT #5, "Date: "; GetSrvTime()
  PRINT #5, "Server: Grumpy/"; pVer
  PRINT #5, "Location: "; RedirectTo
  PRINT #5, "Connection: close"
  PRINT #5, "Allow: GET, HEAD"
  PRINT #5, "Accept-Ranges: none"
  'PRINT #5, "Content-Length: 8700"
  PRINT #5, "Content-Type: text/html; charset=UTF-8"
  PRINT #5,
  IF HeadFlag = 0 THEN PRINT #5, "<html><head><title>301</title></head><body><h1>301 - Moved Permanently</h1><br>The document has moved <a href="""; RedirectTo; """>here</a>.</body></html>"
END SUB


FUNCTION GetSrvTime() AS STRING
  DIM AS STRING Wynik
 'Wynik = FORMAT(NOW, "ddd"","" dd mmm yyyy hh:mm:ss ""GMT""")  ' Old method, using vbcompat.bi
  Dim As Integer res
  Dim As String timestr
  Dim As time_t rawtime
  Dim As tm Ptr timeptr
  'print gmtime
  time_(@rawtime)
  timeptr=gmtime(@rawtime)
  timestr=Space(128)
  res=strftime(Strptr(timestr),128,"%a, %d %b %Y %H:%M:%S GMT",timeptr)
  Wynik = LEFT(timestr,res)
  RETURN Wynik
END FUNCTION


FUNCTION SortStrings(BYVAL StringA AS STRING, BYVAL StringB AS STRING) AS BYTE
  REM  Strings sorting function
  REM   Returns 1 if Strings are in correct order, -1 if strings
  REM   are in the wrong order or zero if StringA = StringB.
  DIM AS INTEGER MaxLen, CharCount
  DIM AS BYTE Wynik = 0
  IF LEN(StringA) > LEN(StringB) THEN
      MaxLen = LEN(StringA)
      StringB += SPACE(MaxLen - LEN(StringB))
    ELSE
      MaxLen = LEN(StringB)
      StringA += SPACE(MaxLen - LEN(StringA))
  END IF
  CharCount = 0
  DO
    CharCount += 1
    IF ASC(MID(StringA, CharCount, 1)) > ASC(MID(StringB, CharCount, 1)) THEN Wynik = 1
    IF ASC(MID(StringA, CharCount, 1)) < ASC(MID(StringB, CharCount, 1)) THEN Wynik = -1
  LOOP UNTIL Wynik <> 0 OR CharCount >= MaxLen
  RETURN Wynik
END FUNCTION


FUNCTION HttpSecurityChecks(Query AS STRING) AS STRING
  REM  This function performs various sanity checks on the given query.
  REM  It returns an empty string if checks are Ok. Otherwise, a string
  REM  containing a human description of the problem will be returned.
  DIM AS STRING Wynik
  DIM AS INTEGER x
  IF INSTR(UCASE(Query), "%2E") > 0 THEN Wynik = "Evasion attempt using twice-encoded '.' character (%2E)."
  IF INSTR(LCASE(Query), "%1u") > 0 THEN Wynik = "Evasion attempt using %1u encoding of the '.' character. Some very old versions of IIS web server understand %1u as '.' character."
  IF INSTR(LCASE(Query), "%u") > 0 THEN Wynik = "The request contains a %u encoding, which is a specific (not standard) Microsoft format used to encode wide characters. This encoding may be used by an attacker to trick systems based on character string recognition."
  IF INSTR(TranslatePercentEnc(Query), "%") > 0 THEN Wynik = "A specific (not supported) %-encoded character has been found, Grumpy doesn't know how to decode it, therefore it could lead to some evasion attempts."
  IF LEN(Query) > 4096 THEN Wynik = "Possible buffer overflow. The request is longer than 4096 bytes. It may be an attempt to insert fraudulent executable code."
  REM Check for non-ASCII characters
  FOR x = 0 TO LEN(Query) - 1
    IF Query[x] > 127 THEN Wynik = "The HTTP request does not contain only 7-bit ASCII characters (the only characters expected in this type of use). Some attackers seek to exploit implementation flaws in certain browsers and web servers by creating specially-crafted HTTP messages to enable executing arbitrary code on the targeted host."
  NEXT x
  REM Check for control characters
  FOR x = 0 TO LEN(Query) - 1
    IF Query[x] < 32 AND Query[x] <> 10 AND Query[x] <> 13 THEN Wynik = "The HTTP request contains some suspect control characters (less than #32 and other than #10 and #13). Some attackers seek to exploit implementation flaws in certain browsers and web servers by creating specially-crafted HTTP messages to enable executing arbitrary code on the targeted host. The detected character is: #" & Query[x] & "."
  NEXT x

  IF INSTR(Query, "./") > 0 THEN Wynik = "An URL containing a dot-slash (""./"") has been detected."
  IF INSTR(Query, "../") > 0 THEN Wynik = "Directory traversal attempt. An URL containing two dots ("".."") has been detected."
  IF INSTR(Query, "\") > 0 THEN Wynik = "A backslash has been detected in the query. There is no reason for that character to appear in a http request."
  RETURN Wynik
END FUNCTION


FUNCTION TranslatePercentEnc(RawString AS STRING, RevEnc AS BYTE = 0) AS STRING
  DIM PercToken(1 TO 19) AS STRING => {"%20","%21","%27","%28","%29","%3A","%40","%26","%3D","%2B","%24","%2C","%23","%5B","%5D","%7E","%2E","%5F","%2D"}
  DIM TextToken(1 TO 19) AS STRING => {" ",  "!",  "'",  "(",  ")",  ":",  "@",  "&",  "=",  "+",  "$",  ",",  "#",  "[", "]",  "~",  ".",  "_", "-"}
  DIM AS STRING Wynik
  DIM AS INTEGER x, StrPointer
  Wynik = RawString
  FOR x = 1 TO 19
    DO
      IF RevEnc = 0 THEN
          StrPointer = INSTR(UCASE(Wynik), PercToken(x))
          IF StrPointer > 0 THEN Wynik = MID(Wynik, 1, StrPointer - 1) + TextToken(x) + MID(Wynik, StrPointer + 3)
        ELSEIF x <= 15 THEN
          StrPointer = INSTR(Wynik, TextToken(x))
          IF StrPointer > 0 THEN Wynik = MID(Wynik, 1, StrPointer - 1) + PercToken(x) + MID(Wynik, StrPointer + 1)
        ELSE
          StrPointer = 0
      END IF
    LOOP UNTIL StrPointer = 0
  NEXT x
  RETURN Wynik
END FUNCTION


SUB About()
  PRINT "Grumpy v"; pVer; " Copyright (C) Mateusz Viste "; pDate
  PRINT
  PRINT "This program is free software: you can redistribute it and/or modify it under"
  PRINT "the terms of the GNU General Public License as published by the Free Software"
  PRINT "Foundation, either version 3 of the License, or (at your option) any later"
  PRINT "version."
  PRINT "This program is distributed in the hope that it will be useful, but WITHOUT ANY"
  PRINT "WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A"
  PRINT "PARTICULAR PURPOSE. See the GNU General Public License for more details."
  PRINT
  PRINT "Grumpy is a simple open-source web server for Linux. Grumpy is not a standalone"
  PRINT "daemon - it requires an inetd-compatible superserver to work. Why one would"
  PRINT "prefer Grumpy over any other Linux web server, like Apache, Lighttp, Jigsaw...?"
  PRINT "Well, there is no reason - I wrote Grumpy for fun. Grumpy doesn't support any"
  PRINT "bells and twistles: the only HTTP commands it knows are ""GET"" and ""HEAD"" (one"
  PRINT "could say it makes it very secure). Grumpy is entirely written in FreeBASIC."
  PRINT
  END
END SUB


FUNCTION CheckForAuth(PathToCheckFor AS STRING) AS STRING
  DIM AS STRING AuthFile, Realm = "", AuthUser, LineBuff
  DIM AS BYTE CharBuff
  DIM AS INTEGER FileHook
  AuthFile = PathToCheckFor
  DO
    CharBuff = AuthFile[LEN(AuthFile) - 1]
    IF CharBuff <> ASC("/") THEN AuthFile = MID(AuthFile, 1, LEN(AuthFile) - 1)
  LOOP UNTIL CharBuff = ASC("/")
  AuthFile += ".grumpy.auth"
  IF FileExists(AuthFile) <> 0 THEN
    IF VerboseMode > 0 THEN PRINT #10, "Auth: The location requires some credentials"
    FileHook = FREEFILE
    OPEN AuthFile FOR INPUT AS FileHook
    IF LOF(FileHook) > 0 THEN
      DO
        LINE INPUT #FileHook, TempString
      LOOP UNTIL EOF(FileHook) OR UCASE(MID(TRIM(TempString), 1, 6)) = "REALM="
      IF UCASE(MID(TRIM(TempString), 1, 6)) = "REALM=" THEN Realm = TRIM(MID(TRIM(TempString), 7))
    END IF
    CLOSE FileHook
    IF Realm = "" THEN Realm = "Authorization required"
  END IF
  IF LEN(Realm) > 0 AND LEN(AuthorizationBuff) > 0 THEN
    AuthUser = MID(AuthorizationBuff, INSTRREV(AuthorizationBuff, " ") + 1)
    FileHook = FREEFILE
    OPEN AuthFile FOR INPUT AS FileHook
    DO
      LINE INPUT #FileHook, LineBuff
      IF LEFT(LCASE(LineBuff), 6) <> "realm=" AND EncBase64(LineBuff) = AuthUser THEN
        Realm = ""
        IF VerboseMode > 0 THEN PRINT #10, "Auth: Access granted ("; AuthUser; ")"
      END IF
    LOOP UNTIL Realm = "" OR EOF(FileHook)
    CLOSE #FileHook
  END IF
  IF Realm <> "" AND VerboseMode > 0 THEN
    IF LEN(AuthUser) > 0 THEN
        PRINT #10, "Auth: Access denied ("; AuthUser; ")"
      ELSE
        PRINT #10, "Auth: Access denied (no credentials provided)"
    END IF
  END IF
  RETURN Realm
END FUNCTION


FUNCTION LineInput(SocketToListenTo AS INTEGER = 0) AS STRING
  DIM AS STRING Wynik, CharBuff
  DIM AS DOUBLE LastData
  DIM AS BYTE TimeoutAbort = 0
  LastData = TIMER
  DO
    CharBuff = INPUT(1, SocketToListenTo)
    IF CharBuff <> CHR(13) AND CharBuff <> CHR(10) AND CharBuff <> "" THEN Wynik += CharBuff: LastData = TIMER
    IF TIMER - LastData > 10 THEN TimeoutAbort = 1
  LOOP UNTIL TimeoutAbort = 1 OR LEN(Wynik) > 1024 * 1024 OR CharBuff = CHR(10)
  IF LEN(Wynik) > 1024 * 1024 THEN
    IF VerboseMode > 0 THEN PRINT #10, "I/O OVERFLOW - Treatement aborted. Dataline to long (>1 MB)."
    CLOSE
    END
  END IF
  IF TimeoutAbort = 1 THEN
    IF VerboseMode > 0 THEN PRINT #10, "I/O TIMEOUT - Treatement aborted. Waiting period to long (>10s)."
    CLOSE
    END
  END IF
  RETURN Wynik
END FUNCTION


FUNCTION GetFileSize(FileToCheck AS STRING) AS STRING
  DIM AS UINTEGER FileSize
  DIM AS INTEGER FileHook
  DIM AS STRING SizeString
  FileHook = FREEFILE
  OPEN FileToCheck FOR BINARY AS FileHook
  FileSize = LOF(FileHook)
  CLOSE #FileHook
  IF FileSize >= 1024^3 THEN
      SizeString = FORMAT(FileSize / 1024^3, "#.0") & " GiB"
    ELSEIF FileSize >= 1024^2 THEN
      SizeString = FORMAT(FileSize / 1024^2, "#.0") & " MiB"
    ELSEIF FileSize >= 1024 THEN
      SizeString = FORMAT(FileSize / 1024, "#.0") & " KiB"
    ELSE
      SizeString = FileSize & " bytes"
  END IF
  RETURN SizeString
END FUNCTION


FUNCTION GetFileDate(FileToCheck AS STRING) AS STRING
  DIM MonthList(1 TO 12) AS STRING => {"Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"}
  DIM AS INTEGER fDay, fMonth, fYear, fHour, fMinute
  DIM AS DOUBLE fTimeSerial
  DIM AS STRING Wynik
  fTimeSerial = FileDateTime(FileToCheck)
  fDay = DAY(fTimeSerial)
  fMonth = MONTH(fTimeSerial)
  fYear = YEAR(fTimeSerial)
  fHour = HOUR(fTimeSerial)
  fMinute = MINUTE(fTimeSerial)
  Wynik = "" & fDay & " " & MonthList(fMonth) & " " & fYear & ", " & fHour & ":" & RIGHT("0" & fMinute, 2)
  RETURN Wynik
END FUNCTION


FUNCTION GetFileDescription(FileToCheck AS STRING, DescrPath AS STRING) AS STRING
  DIM AS STRING Wynik, DescrFile, Token1, Token2, LineBuff
  DIM AS BYTE CharBuff
  DIM AS INTEGER FileHook
  DescrFile = DescrPath + "descript.ion"
  FileHook = FREEFILE
  IF FileExists(DescrFile) <> 0 THEN
    OPEN DescrFile FOR INPUT AS FileHook
    IF LOF(FileHook) > 0 THEN
      DO
        LINE INPUT #FileHook, LineBuff
        LineBuff = TRIM(LineBuff)
        Token1 = TRIM(MID(LineBuff, 1, INSTR(LineBuff, CHR(9)) - 1))
        Token2 = TRIM(MID(LineBuff, INSTR(LineBuff, CHR(9)) + 1))
        IF Token1 = FileToCheck THEN Wynik = Token2
      LOOP UNTIL EOF(FileHook) OR LEN(Wynik) > 0
    END IF
    CLOSE FileHook
  END IF
  RETURN Wynik
END FUNCTION